  /********************************************************************
   *                                                                  *
   *  Project: 16F1828 Clock v1                                       *
   *   Source: 16F1828_Clock_v1.c                                     *
   *   Author: Mike McLaren, K8LH                                     *
   *  (C)2012: Micro Application Consultants                          *
   *     Date: 02-Nov-2012                                            *
   *                                                                  *
   *   16F1828 "4-Digit Clock" experiment (lighting one segment at    *
   *   a time at 1/32nd duty cycle on the K8LH 4-Digit Demo Board).   *
   *                                                                  *
   *      IDE: MPLAB 8.84 (tabs = 4)                                  *
   *     Lang: SourceBoost BoostC v7.05, Lite/Free version            *
   *                                                                  *
   ********************************************************************/

   #include <system.h>

   #pragma DATA _CONFIG1, _FOSC_INTOSC & _WDTE_OFF & _MCLRE_OFF
   #pragma DATA _CONFIG2, _PLLEN_OFF & _LVP_OFF

   #pragma CLOCK_FREQ 8000000       // 8 MHz INTOSC

  /********************************************************************
   *  function prototypes                                             *
   ********************************************************************/

  /********************************************************************
   *  typedef and defines                                             *
   ********************************************************************/

   typedef unsigned char u08;
   typedef unsigned int u16;

   #define r08 const rom unsigned char

   #define piezo porta.1            // piezo speaker (RA1)

  /********************************************************************
   *  variables (and const arrays)                                    *
   ********************************************************************/

   u08 colsel = 0x10;               // column select mask
   u08 segmask = 0x01;              // segment select mask
   u08 segdata;                     //
   u08 group = 0;                   // set mode flash group mask

   u08 secs = 0x00;                 // packed bcd seconds
   u08 mins = 0x00;                 // packed bcd minutes
   u08 hour = 0x00;                 // packed bcd hours

   u08 swnew = 0;                   // switch state logic work var
   u08 swold = 0;                   // switch state latch
   u08 flags = 0;                   // switch flags (bits 7..4)
   u08 beepctr = 0;                 // beep task var

   r08 segment[] = { 0b00111111,    // "0"   -|-|F|E|D|C|B|A
                     0b00000110,    // "1"   -|-|-|-|-|C|B|-
                     0b01011011,    // "2"   -|G|-|E|D|-|B|A
                     0b01001111,    // "3"   -|G|-|-|D|C|B|A
                     0b01100110,    // "4"   -|G|F|-|-|C|B|-
                     0b01101101,    // "5"   -|G|F|-|D|C|-|A
                     0b01111101,    // "6"   -|G|F|E|D|C|-|A
                     0b00000111,    // "7"   -|-|-|-|-|C|B|A
                     0b01111111,    // "8"   -|G|F|E|D|C|B|A
                     0b01101111,    // "9"   -|G|F|-|D|C|B|A
                     0b00000000,    // " "   -|-|-|-|-|-|-|-
                     0b01000000 };  // "-"   -|G|-|-|-|-|-|-


  /********************************************************************
   *  functions                                                       *
   ********************************************************************/

  /********************************************************************
   *  main init                                                       *
   ********************************************************************/

   void main()
   { ansela = 0;                //
     anselb = 0;                //
     anselc = 0;                //
     porta = 0;                 //
     portb = 0;                 //
     portc = 0;                 //
     trisa = 0;                 //
     trisb = 0;                 //
     trisc = 0;                 //
     osccon = 0b01110000;       // set INTOSC to 8 MHz
     while(!oscstat.HFIOFS);    // wait for oscillator stable
  /*                                                                  *
   *  config TMR1 for low power 32768 Hz oscillator operation         *
   *                                                                  */
     t1con = 0b10001101;        // 10------ TMR1CS source T1OSC
                                // --00---- T1CKPS prescale 1
                                // ----1--- T1OSCEN osc enabled
                                // -----1-- T1SYNC asynchronous
                                // -------1 TMR1ON timer 'on'
  /*                                                                  *
   *  250-us (500 cycle) TMR2 interrupts for a 125-Hz refresh rate    *
   *                                                                  */
     tmr2 = 0;                  // clear Timer 2 register
     t2con = 0b00000001;        // '0-------' unimplemented bit
                                // '-0000---' TOUTPS<3:0>, post 1
                                // '-----0--' TMR2ON, Timer 2 off
                                // '------01' T2CKPS<1:0>, pre 4
     pr2 = 125-1;               // 125 x 2-usec 'ticks' = 250 usecs
     pir1 = 0;                  // clear peripheral interrupt flags
     pie1 = 1<<TMR2IE;          // set Timer 2 interrupt enable bit
     intcon = 0b11000000;       // '1-------' GIE, enable global and
                                // '-1------' PEIE, peripheral ints
                                // '--0-----' T0IE, TMR0 ints off
                                // '---0----' INTE, off
                                // '----0---' GPIE, IOC disabled
                                // '-----000' T0IF/INTF/GPIF flags
     t2con.TMR2ON = 1;          // start TMR2 (and interrupts)

  /********************************************************************
   *  main loop                                                       *
   ********************************************************************/

     while(1)                   //
     { if(flags.4)              // if set switch "on"
       { group = 0b00110000;    // flash "hour" display group
         while(flags.4)         // while in set mode
         { 
          /*                                                          *
           *  rt arrow selects between "hour" & "mins" groups         *
           *                                                          */
           if(flags.7)          // if rt arrow
           { flags.7 = 0;       // clear switch flag and
             group ^= 0xF0;     // toggle flash group
           }                    //

          /*                                                          *
           *  up arrow increments the packed bcd group value          *
           *                                                          */
           if(swold.5)          // if up arrow
           { if(group.4)        // if "hour" group
               if(hour++ == 0x23) hour = 0x00;
             if((hour & 15) == 10) hour += 6;

             if(group.6)        // if "mins" group
               if(mins++ == 0x59) mins = 0x00;
             if((mins & 15) == 10) mins += 6;
             delay_ms(200);     // repeat interval
           }                    //

          /*                                                          *
           *  dn arrow decrements the packed bcd group value          *
           *                                                          */
           if(swold.6)          // if dn arrow
           { if(group.4)        // if "hour" group
               if(hour-- == 0x00) hour = 0x23;
             if((hour & 15) == 15) hour -= 6;

             if(group.6)        // if "mins" group
               if(mins-- == 0x00) mins = 0x59;
             if((mins & 15) == 15) mins -= 6;
             delay_ms(200);     // repeat interval
           }                    //
         }                      //
         group = 0;             // flash off
         flags = 0;             // clear switch flags
         tmr1l = 0;             //
         tmr1h = 128;           //
       }                        //
     }
   }

  /********************************************************************
   *  interrupt                                                       *
   ********************************************************************/

   void interrupt()             // 250-us (500 cyc) TMR2 interrupts
   { u08 index;                 //
     pir1.TMR2IF = 0;           // clear TMR2 interrupt flag
    /*                                                                *
     *  refresh display (display new segment)                         *
     *                                                                */
     portb = 0;                 // blank the display
     portc = ~segdata;          // new segment (invert, active lo)
     portb = colsel;            // light new segment (column "on")

    /*                                                                *
     *  beep task generates a 32-ms 2000-Hz "new press" beep          *
     *                                                                */
     if(beepctr)                // if beep task running
     { piezo ^= 1;              // toggle piezo spkr pin 
       beepctr--;               //
     }                          //

    /*                                                                *
     *  sample one new switch each 2-mS "new column" interval for     *
     *  an effective 8-mS debounce interval per switch.               *
     *                                                                */
     if(segmask.0)              // if segment A (2-mS) interval
     { swnew = swold | colsel;  // indicate pressed unconditionally
       if(porta.3 == 0)         // if not pressed (RA3 = 0)
         swnew ^= colsel;       // indicate released
    /*                                                                *
     *  K8LH Parallel Switch State Logic (with "new press" filter)    *
     *                                                                *
     *  swnew  ___---___---______  sample active hi switches          *
     *  swold  ____---___---_____  switch state latch                 *
     *  swnew  ___-__-__-__-_____  changes, press or release          *
     *  swnew  ___-_____-________  filter out 'release' bits          *
     *  flags  ___-------________  toggle flag bits for main          *
     *                                                                */
       swnew ^= swold;          // changes, press or release
       swold ^= swnew;          // update switch state latch
       swnew &= swold;          // filter out 'release' bits
       if(swnew)                // if any "new press" bits
         beepctr = 128;         // task a "new press" beep
       flags ^= swnew;          // toggle flag bits for main
     }                          //

    /*                                                                *
     *  advance segment mask bit for next interrupt, and if first     *
     *  segment, bump the column select bit mask.                     *
     *                                                                */
     segmask <<= 1;             // 
     if(segmask == 0)           // if last segment (2-mS interval)
     { segmask = 1;             // reset to segment A and
       colsel <<= 1;            // advance column select bit
       if(colsel == 0)          // if last column
         colsel = 0x10;         // reset to first column (RB4)
     }                          //

    /*                                                                *
     *  prepare new segment data for next interrupt                   *
     *                                                                */
     index = hour >> 4;         // 0..9
     if(!index) index = 10;     // suppress a leading zero
     if(colsel.5) index = hour & 15;
     if(colsel.6) index = mins >> 4;
     if(colsel.7) index = mins & 15;
     segdata = segment[index];  //
     if(colsel.4)               // if column/digit 1
       if(tmr1h.6)              // if 1/2 second interval
         segdata.7 = 1;         // turn colon (RB7) on
     segdata &= segmask;        // only one segment each interrupt

    /*                                                                *
     *  flash hour or mins group at 2-Hz rate when in set mode        *
     *                                                                */
     if(flags.4)                // if SET switch "on" and
     { if(colsel & group)       // if SET display group and
       { if(tmr1h.5)            // if 250-mS interval
           segdata &= 0x80;     // turn the segment off
       }                        //
       secs = 0;                // 
     }                          //

    /*                                                                *
     *  real-time-clock                                               *
     *                                                                */
     if(pir1.TMR1IF)            // if TMR1 overflow flag
     { pir1.TMR1IF = 0;         // clear the flag and
       tmr1h.7 = 1;             // reset TMR1 for 1-Second
       if(secs++ == 0x59)       // if 'seconds' rollover
       { secs = 0x00;           // set 'seconds' to 0x00 and
         if(mins++ == 0x59)     // if 'minutes' rollover
         { mins = 0x00;         // set 'minutes' to 0x00 and
           if(hour++ == 0x23)   // if 'hour' rollover
             hour = 0x00;       // reset 'hour' to 0x00
         }                      //
       }                        //
       if((secs & 15) == 10)    // 
         secs += 6;             // bcd adjust
       if((mins & 15) == 10)    //
         mins += 6;             // bcd adjust
       if((hour & 15) == 10)    //
         hour += 6;             // bcd adjust
     }                          //
   }                            //

